/*
 *		gssub.exe:	Utility program for dviout/dviprt
 *
 *		Make "gs_exec.bat" to make *.pbm/*.gif from *.ps 
 *
 *									written by SHIMA	   Feb 21, 1993
 *                                  modified by Asayama    Apr  2, 1993
 *                                  rewritten by Asayama   May  1, 1994
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
  
#define READ_BUF_SIZE 1024
  
  /*
    %GSS%DefineDevice: <device driver> <xres> <yres>
    
    %GSS%BeginProlog:
    controle procedures
    %GSS%EndProlog
    
    %GSS%BeginUnit: <file name> <dvi page> <count>
    page setup
    %GSS%BeginContents:
    ...
    %GSS%EndContents
    %GSS%EndUnit
    */
  
char *GSSComment = "%GSS%";
char *DefineDevice = "%GSS%DefineDevice:";
char *BeginProlog = "%GSS%BeginProlog:";
char *EndProlog = "%GSS%EndProlog";
char *BeginUnit = "%GSS%BeginUnit:";
char *EndUnit = "%GSS%EndUnit";
char *BeginCont = "%GSS%BeginContents";
char *EndCont = "%GSS%EndContents";

char *InlinePSFile = "tmp$$.ps";
char *EPSFile = "tmp$.ps";

char ghostscript[128] = "gs";
char gs_args[256] = "";

void help(int );
void CannotOpen(char *);
void eputs(char *);
void eprintf(char *,...);
void exit_process(void);

FILE *fplist;
double x_resolution = 72.0,y_resolution = 72.0;
char read_buf[READ_BUF_SIZE];
char sub_buf[READ_BUF_SIZE];
int xsize = 0,ysize = 0;
char *OutputFile = NULL;
char *InputFile = NULL;
char *Device = NULL;
int inline_f = 0;

#define NO_FILE -2
#define NOT_FOUND -1
#define FOUND_SAME 0

int main(int argc,char *argv[])
{
  int code;
  char *env;
  char *BatchFile = "gs_exec.bat";
  char *PSFile = "gs_exec$.ps";
  int page,count = 0;
  int units;
  
  argc = ScanArgs(argc,argv);
  
  inline_f = !strcmp(InlinePSFile,InputFile);
  if (!inline_f) {
    if (strcmp(EPSFile,InputFile)) {
      eprintf("Invalid input file name `%s'.\n",InputFile);
      eprintf("`%s' or `%s' required.\n",EPSFile,InlinePSFile);
      exit(1);
    }
  }
  if (xsize == 0 && !inline_f) {
    eputs("Size of the output image is not specified.\n");
    exit(1);
  }
  
  env = getenv("page");
  if (env == NULL) {
    eputs("dviout/dviprt seems old version.\n");
    exit(1);
  }
  page = atoi(env);
  env = getenv("gsx");
  if (env != NULL && strlen(env) <= 120){
    strcpy(ghostscript, env);
    for(env = ghostscript; *env; env++){
      if(*env == '^') *env = ' ';
    }
  }
  
  atexit(exit_process);
  
  fplist = tmpfile();
  if (fplist == NULL) CannotOpen("temporary file");
  count = ScanOutputFile(PSFile,InputFile,OutputFile,&units);
  if (count == 0) {
    eprintf("The same codes are found in the output file `%s'. Skip.\n",
	    PSFile);
  }
  else {
    if (count > 0) {
      eprintf("WARNING The same PS file is used in a different way.\n");
      count++;
    }
    else if (count == NO_FILE) {
      WriteHeader(PSFile,Device,x_resolution,y_resolution);
      count = 1;
    }
    else if (count == NOT_FOUND) count = 1;
    
    units++;
    fprintf(fplist,"%s -g%dx%d",OutputFile,xsize,ysize);
    AddPS(InputFile,PSFile,OutputFile,page,count,xsize,ysize);
    eprintf("Wrote on `%s'.\n",PSFile);
    eprintf("Execute `%s' to make `%s'.\n",BatchFile,OutputFile);
  }
  fseek(fplist,0,0);
  MakeBatch(BatchFile,PSFile,units);
  
  return 0;
}

void RemoveLF(char *buf)
{
  int i;
  char *pbuf = buf+strlen(buf)-1;
  while (pbuf >= buf && isspace(*pbuf)) pbuf--;
  *++pbuf = 0;
}

int NextUnit(FILE *fp)
{
  while (fgets(read_buf,READ_BUF_SIZE,fp)) {
    RemoveLF(read_buf);
    if (strcmp(read_buf,EndUnit) == 0)
      break;
  }
  return 0;
}

int ScanOutputFile(char *fname,char *ifile,char *ofile,int *units)
{
  FILE *fp;
  int ps_start = 0;
  char target_name[128];
  int page,count,x_size,y_size;
  int gssc_off = strlen(GSSComment);
  int same_f = 0;
  int max_count = 0;
  int found_f = 0;
  
  *units = 0;
  
  if (access(fname,0)) return NO_FILE;

  fp = fopen(fname,"r");
  if (fp == NULL) CannotOpen(fname);
  
  if (fgets(read_buf,READ_BUF_SIZE,fp) == NULL) {
    fclose(fp);
    return NO_FILE;
  }
  if (strncmp(read_buf,DefineDevice,strlen(DefineDevice)) == 0) {
    double rx,ry;
    int code = sscanf(read_buf+strlen(DefineDevice),
		      "%s %lf %lf",sub_buf,&rx,&ry);
    if (code != 3) {
    not_gssub_output:
      eprintf("File %s is not output of gssub.\n",fname);
      exit(1);
    }
    if (rx != x_resolution || ry != y_resolution) {
      eputs("Resolution is not match.\n");
      exit(1);
    }
  }
  else goto not_gssub_output;
  
  while (fgets(read_buf,READ_BUF_SIZE,fp) != NULL) {
    if (strncmp(read_buf,GSSComment,gssc_off) == 0) {
      RemoveLF(read_buf);
      if (strncmp(BeginUnit,read_buf,strlen(BeginUnit)) == 0) {
	int code;
	
	if (ps_start) goto not_gssub_output;
	
	ps_start = 1;
	code = sscanf(read_buf+strlen(BeginUnit),
		      "%s %d %d %d %d",target_name,&page,&count,&x_size,&y_size);
	if (code != 5) goto not_gssub_output;
	same_f = !strcmp(target_name,ofile);
	(*units)++;
    fprintf(fplist,"%s -g%dx%d\n",target_name,x_size,y_size);
      }
      else if (strcmp(EndUnit,read_buf) == 0) {
	if (ps_start == 0)
	  goto not_gssub_output;
	ps_start = 0;
	same_f = 0;
      }
      else if (ps_start) {
	if (strcmp(BeginCont,read_buf) == 0) {
	  FILE *fps;
	  if (same_f) {
        found_f = 1;
	    fps = fopen(ifile,"r");
	    if(fps == NULL) CannotOpen(ifile);
	    while (fgets(read_buf,READ_BUF_SIZE,fp)) {
	      RemoveLF(read_buf);
	      if (strcmp(read_buf,EndCont) == 0) {
		if (fgetc(fps) == EOF) break; /* same */
		else {
		found_but_diff:
		  if (count > max_count) max_count = count;
		  NextUnit(fp);
		  ps_start = same_f = 0;
		  break;
		}
	      }
	      if (fgets(sub_buf,READ_BUF_SIZE,fps) == NULL)
		goto found_but_diff;
	      RemoveLF(sub_buf);
	      if (strcmp(read_buf,sub_buf))
		goto found_but_diff;
	    }
	    fclose(fps);
	  }
	  else { NextUnit(fp); ps_start = same_f = 0;}
	}
      }
    }
  }
  
  fclose(fp);
  return found_f ? max_count : NOT_FOUND;
}

int WriteHeader(char *fname,char *dev,double xres,double yres)
{
  FILE *fp;
  char *check[] = {
    "systemdict /DVIPAGE 2 copy known",
    "{get type /integertype ne",
    "{(Invalid page number.\\n) print flush 1 .quit} if}  {pop pop} ifelse",
    NULL
  };
  char *defs[] = {
    "20 dict begin",
    "/@UnitNumber 0 def",
    "currentdevice getdeviceprops",
	"counttomark 2 idiv dup dict begin {def} repeat pop currentdict end",
    "dup /HWSize get /@DefaultSize exch def",
    "pop",
    "/@CheckUnit {",
    "  /@UnitNumber @UnitNumber 1 add def", 
    "  /@DeviceSize exch def /@Count exch def",
    "  /@Page exch def /@OutputFile exch def",
    "  systemdict /DVIPAGE 2 copy known",
    "    { get revision 250 ge { @Page } { @UnitNumber } ifelse eq }",
    "    { pop pop @Count 1 eq } ifelse",
    "} bind def",
    "/@SetDevice {",
    "  mark /HWSize @DeviceSize /OutputFile @OutputFile",
    "  currentdevice putdeviceprops setdevice",
    "} bind def",
    "/@BeginGSSEPSF { userdict begin mark save } bind def",
    "/@EndGSSEPSF {",
    "  counttomark 1 sub dup 0 gt { (\\nGSSUB: Warning  ) print",
    "    dup ==only ( object\\(s\\) remain on operand stack.\\n) print flush",
    "    {pop} repeat } {pop}ifelse restore end",
    "} bind def",
    "/@BeginGSSILPS {} bind def",
    "/@EndGSSILPS {} bind def",
    "/@BeginGSSUB {",
    "  (Making `) print @OutputFile print",
    "  (' in the page ) print @Page ==only (... ) print flush",
    "  revision 250 ge { @SetDevice } if",
    "} bind def",
    "/@EndGSSUB {",
    "  (done.\\n) print flush",
    "  revision 250 lt { @UnitNumber 1 eq {255 .quit} { quit } ifelse } if",
    "} bind def",
    "/quit {} bind def",
    NULL};
  
  fp = fopen(fname,"w");
  if (fp == NULL) CannotOpen(fname);
  
  fprintf(fp,"%s%s %d %d\n",DefineDevice,dev,(int)xres,(int)yres);
  fprintf(fp,"%s\n",BeginProlog);
  WriteStringsList(check,fp);
  WriteStringsList(defs,fp);
  fprintf(fp,"%s\n",EndProlog);
  
  fclose(fp);
  return 0;
}

int AddPS(char *ifile,char *ofile,char *target,int page,int count,int w,int h)
{
  FILE *ifp,*ofp;
  
  ifp = fopen(ifile,"r");
  if (ifp == NULL) CannotOpen(ifile);
  ofp = fopen(ofile,"a");
  if (ofp == NULL) CannotOpen(ofile);
  
  fprintf(ofp,"\n%s%s %d %d %d %d\n",BeginUnit,target,page,count,w,h);
  fprintf(ofp,"(%s) %d %d ",target,page,count);
  if (w == 0) fputs("@DefaultSize",ofp);
  else fprintf(ofp,"[%d %d]",w,h);
  fputs(" @CheckUnit {\n",ofp);
  fprintf(ofp,"@BeginGSSUB\n");
  fputs(inline_f ? "@BeginGSSILPS\n" : "@BeginGSSEPSF\n",ofp);
  fprintf(ofp,"%s\n",BeginCont);
  while (fgets(read_buf,READ_BUF_SIZE,ifp))
    fputs(read_buf,ofp);
  fprintf(ofp,"%s\n",EndCont);
  fputs(inline_f ? "@EndGSSILPS\n" : "@EndGSSEPSF\n",ofp);
  fprintf(ofp,"@EndGSSUB } if\n");
  fprintf(ofp,"%s\n",EndUnit);
  
  fclose(ifp);
  fclose(ofp);
  return 0;
}

int WriteStringsList(char **p,FILE *fp)
{
  while (*p) {
    fputs(*p,fp);
    fputs("\n",fp);
    p++;
  }
  return 0;
}

int MakeBatch(char *fname,char *psfile,int units)
{
  int i;
  FILE *fp;
  char *b1[] = {
    "echo off",
    "if \"%1\"==\"\" goto make_all",
    "if not \"%2\"==\"\" goto arg_err",
    NULL,
  };
  char *b2[] = {
    "if errorlevel 255 goto oldver",
    "goto bye",
    ":oldver",
    NULL,
  };
  char *b3[] = {
    "goto bye",
    ":arg_err",
    "echo Too many arguments.>con",
    ":bye",
    NULL,
  };
  
  fp = fopen(fname,"w");
  if (fp == NULL) CannotOpen(fname);
  
  WriteStringsList(b1,fp);
  fputs(ghostscript,fp);
  fputs(" -dDVIPAGE=%1",fp);
  fputs(gs_args,fp);
  fprintf(fp," %s\n",psfile);
  fputs("goto bye\n",fp);
  
  fputs(":make_all\n",fp);
  fgets(read_buf,READ_BUF_SIZE,fplist);
  RemoveLF(read_buf);
  fprintf(fp,"%s -sOUTPUTFILE=%s %s %s\n",ghostscript,read_buf,gs_args,psfile);
  WriteStringsList(b2,fp);

  for (i=0;i<units-1;i++) {
   	fgets(read_buf,READ_BUF_SIZE,fplist);
    RemoveLF(read_buf);
    fprintf(fp,"%s -sOUTPUTFILE=%s -dDVIPAGE=%d%s %s\n",
		ghostscript,read_buf,i+2,gs_args,psfile);
  }
  
  WriteStringsList(b3,fp);
  fclose(fp);
  
  return 0;
}

void ProcArg(char *arg)
{
  int i,code;
  double rvalue1,rvalue2;
  int ivalue1,ivalue2;
  char *pstr;
  typedef struct {
    char *name;
    char **ptr;
    int flag;
  } name_t;
  name_t *pname;
  name_t key[] = {
    { "OutputFile",&OutputFile,0 },
    { "OUTPUTFILE",&OutputFile,0 },
    { "DEVICE", &Device,1 },
    { NULL },
  };

  if (*arg == '-') {
    switch (*(arg + 1)) {
      case 'g': /* size of output image */
        code = sscanf(arg + 2,"%dx%d",&ivalue1,&ivalue2);
        if (code != 2) {
          eputs("-g<xsize>x<ysize> required\n");
          exit(1);
        }
        if (ivalue1 <= 0 || ivalue2 <= 0) {
          eprintf("Size %s is out of range.\n",arg);
          exit(1);
        }
        xsize = ivalue1;
        ysize = ivalue2;
        break;
      case 'r': /* resolution */
        code = sscanf(arg + 2,"%lfx%lf",&rvalue1,&rvalue2);
        if (code < 1) {
          eputs("-r<xres>x<yres> or -r<res> required\n");
          exit(1);
        }
        x_resolution = rvalue1;
        y_resolution = (code == 1) ? rvalue1 : rvalue2;
        goto raw_args;
      case 's': /* dictionary */
        pstr = strchr(arg + 2, '=');
        if (pstr == NULL) {
          eputs("-s<name>=<strings> required.\n");
          exit(1);
        }
        ivalue1 = pstr - (arg + 2);
        pstr++;
        for (pname = key;pname->name;pname++) {
          if (strlen(pname->name) == ivalue1 &&
            strncmp(pname->name,arg + 2,ivalue1) == 0) {
            *(pname->ptr) = malloc(strlen(pstr) + 1);
            strcpy(*(pname->ptr), pstr);
            if (pname->flag) goto raw_args;
            break;
          }
        }
        if (pname->name == NULL)  goto raw_args; /* not found */
        break;
      default: /* other options */
        goto raw_args;
    }
  }
  else {
raw_args:
    InputFile = gs_args + strlen(gs_args);
    strcat(gs_args," ");
    strcat(gs_args,arg);
  }
}

int ScanArgs(int argc,char *argv[])
{
  int i;
  char *ptr1, *ptr2;
  FILE *fp;
  
  if (argc < 2) help(1);
  
  InputFile = NULL;
  for (i=1;i<argc;i++) {
    if (*argv[i] == '@') {
      if ((fp = fopen(argv[i] + 1, "r")) == NULL) {
        CannotOpen(argv[i] + 1);
      }
      while (fgets(read_buf,READ_BUF_SIZE,fp)) {
        ptr1 = ptr2 = read_buf;
        while (*++ptr2) {
          if (isspace(*ptr2)) {
            *ptr2 = '\0';
            if (*ptr1) ProcArg(ptr1);
            ptr1 = ptr2 + 1;
          }
        }
        if (*ptr1) ProcArg(ptr1);
      }
      fclose(fp);
    }
    else ProcArg(argv[i]);
  }

  if (InputFile == NULL) {
    eputs("Too few argument(s).");
    exit(1);
  }
  *InputFile = '\0';
  InputFile++;
  
  if (OutputFile == NULL) {
    eputs("The name of output file is not specified.\n");
    exit(1);
  }
  return 0;
}

void help(int code)
{
  fprintf(stderr,"gssub: Used by dviout/dviprt with -GS=5 option\n"
        "\t\tVer 0.4 written by SHIMA, Mar 7 1993\n"
		"\t\tVer 0.5 rewritten by K.Asayama, May 1994\n"
		"\t\tVer 0.6 modified by K.Matsuda, April 1995\n");
  exit(code);
}

void CannotOpen(char *fname)
{
  eprintf("Cannot open file `%s'.\n",fname);
  exit(1);
}

void eputs(char *str)
{
  fputs("GSSUB: ",stderr);
  fputs(str,stderr);
}

void eprintf(char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  fputs("GSSUB: ",stderr);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
}

void exit_process(void)
{
  exit(1);
}
