
/*
 *  TEX Device Driver  ver 2.02-
 *  copyright(c) 1988, 1989 by TSG, 1990-93 by SHIMA
 *
 *  loadpk.c :
 *            April 15, 1989 : first edition
 *                             programed by T.IWAI
 *  1991 changed by hideki
 *	very slightly modified against warnings by Oh-Yeah? 25 May 1992
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <errno.h>
#include <string.h>
#include "dd.h"
#include "dviread.h"
#if defined(TURBOC) || defined(MSC)
#ifndef _DEF_DOS_H_
#include <dos.h>
#define _DEF_DOS_H_
#endif
#endif
#include "err.h"

#ifndef min
#  define min(a,b) ( (a) < (b) ? (a) : (b) )
#endif
#ifndef max
#  define max(a,b) ( (a) > (b) ? (a) : (b) )
#endif

#if DEBUG & MEM_CHK
static void (*_bufchk) ();

#endif

/* buffer.c */
void pk_buf_flush(void);

/* loadpk.c */
char *to_font_name(char *);

/* option.c */
int make_path(char *, char *, char *, char *);

int file_read(int, BUFFER *, int);
int ask_pk_buf(long size);
int openf(char *, int);
int flush_pk(void);
int flush_jpk(void);
#ifdef TTFONT
int load_tfm(char *);
#endif

int f_flush_pk;

/* vfont.c */
void flush_vfn(void);

#ifdef TTFONT
#include "ttfont.h"
/* ttfont.c */
uchar *get_cmttinfo(int);
#endif

#define READ_SIZE (1024*8)
/*  xɓǂݍރTCY */

static void buf_load(int fd, BUFFER *buf, long size)
	/*  bufsizeǂݍ݂܂DrŃt@CIĂ
     *  G[ɂ͂Ȃ܂Dt@CN[Y܂D
     */
{
	int tmp_size;

	ENTER("buf_load");

	while (!eof(fd) && size > 0) {
		tmp_size = (int)min(READ_SIZE, size);
#if DEBUG & MEM_CHK
		(*_bufchk) (buf);
#endif
		if (file_read(fd, buf, tmp_size) == FAILURE)
			error(FILE_FAULT, "fail to read FONT");

		size -= tmp_size;
		buf += tmp_size;
	}
	END();
}

FILE *fopenf(char *name, char *mode)
{
	FILE *fp;

	if (*mode != 'w' && access(name, 0) != 0) return( NULL );
	if ((fp = fopen(name, mode)) != NULL)
		return( fp );
	if( flush_pk() ){
		if ((fp = fopen(name, mode)) != NULL)
			return( fp );
	}
	if( flush_jpk() ){
		if ((fp = fopen(name, mode)) != NULL)
			return( fp );
	}
#ifndef	NOVFONT
	flush_vfn();
#endif
	return( fopen(name, mode) );
}

int openf(char *name, int mode)
{
	int fd;

	if ((fd = open(name, mode)) > 0 || errno != EMFILE )
		return (fd);
	if( flush_pk() ){
		if ((fd = open(name, mode)) > 0 || errno != EMFILE )
			return (fd);
	}
	if( flush_jpk() ){
		if ((fd = open(name, mode)) > 0 || errno != EMFILE )
			return (fd);
	}
#ifndef	NOVFONT
	flush_vfn();
#endif
	return( open(name,mode) );
}

int flush_pk()
{
	static FONT_INFO *font;
	FONT_INFO *fontc;

	if (font == NULL)
		font = first_font_info;
	fontc = font;
	do {
		if ((fontc->font_type == PKD_FONT
#ifdef VFD
			|| fontc->font_type == VFD_FONT
#endif
#ifdef TTFONT
			|| fontc->font_type == CMTTF
#endif
			|| fontc->font_type == JXL)
			&& fontc->ext.fh > 0) {
			close(fontc->ext.fh);
			f_flush_pk++;
			fontc->ext.fh = 0;
			font = font->next_font;
			return( 1 );
		}
		fontc = fontc->next_font;
		if (fontc == NULL)
			fontc = first_font_info;
	} while (font != fontc);
	return( 0 );
}


int flush_jpk()
{
	static FONT_INFO *font;
	static int flush_id;
	int fh;
	FONT_INFO *fontc;

	if (font == NULL)
		font = first_font_info;
	fontc = font;
	do {
		if ((fontc->font_type >= JIS_FONT)
			&& fontc->ext.kdir->fh > 0) {
			close(fontc->ext.kdir->fh);
			f_flush_pk++;
			flush_id--;
			fh = fontc->ext.kdir->fh;
			font = font->next_font;
			for( fontc = first_font_info; fontc; fontc = fontc->next_font ){
				if(fontc->font_type >= JIS_FONT && fontc->ext.kdir->fh == fh)
					fontc->ext.kdir->fh = flush_id;
			}
			return ( 1 );
		}
		fontc = fontc->next_font;
		if (fontc == NULL)
			fontc = first_font_info;
	} while (font != fontc);
	return( 0 );
}


void re_openf(FONT_INFO *font)
{
	int id, fh;
	FONT_INFO *fontc;

	id = font->ext.fh;
	if ( (fh = openf(font->ext.kdir->name, O_BINARY|O_RDONLY)) < 0 )
		error(FILE_FAULT, "Cannot re_open FONT file %s", font->ext.kdir->name);
	for( fontc = first_font_info; fontc; fontc = fontc->next_font ){
		if( fontc->ext.fh == id ) font->ext.fh = fh;
	}
}

#if defined(BC31) && (__TURBOC__ < 0x452)
#pragma option -3-
#endif

int ask_pk_buf(long size)
{
	if ((HUGE_BUF *)pk_buf_pointer->end -
		(HUGE_BUF *)pk_buf_pointer->current <= size + 32) {
		if (pk_buf_pointer->size < size || size > 0xffe0L)
			return (FAILURE);
		pk_buf_flush();
	}
	return (NOTHING);
}

#ifdef BC31
#pragma option -3
#endif

static BUFFER *
    load_pk_data(char *name)
	/*  load_pk̃XPgW[D namẽtHgt@C
     *  ꊇ bufɓǂݍށD obt@Ƃ́Cflush
     *  ĂяoD
     *  Ԓl͊i[ꏊD
     */
{
	int fd, i;
	long file_pos, file_size;
	long far *jxl_info;
	long tmp;
	FILE *fp;
	BUFFER *pk;
#ifdef TTFONT
	uchar *s;
	BUFFER *d;
#endif

#define	JXL4_ID		(0x258B0100L)
#define	CMTTF_ID	('t' + 't' + 'f')

	ENTER("load_pk_data");

	if (*name == '*') {
		file_pos = *((long *)(name + 1));
		file_size = *((long *)(name + 5));
		if ((fd = openf(name + 9, O_RDONLY | O_BINARY)) == FAILURE
			|| lseek(fd, file_pos, SEEK_SET) == -1L)
			error(NO_FILE, name+9);
	}
	else {
		if ((fd = openf(name, O_RDONLY | O_BINARY)) == FAILURE)
			error(NO_FILE, name);
		file_size = filelength(fd);
	}
	_read(fd, &tmp, 4);
#ifdef TTFONT
	/* gq .ttf Ȃ TFM ǂݍ */
	if ((i = strlen(name) - 4) > 0 && strcmpi(name + i, ".ttf") == 0) {
		if ((s = get_cmttinfo(fd)) == NULL)
			error(FILE_FAULT, "Invalid TrueType font %s", name);
		close(fd);
		file_size = sizeof(CMTT_INFO) + 4;
		if ((fd = load_tfm(name)) > 0) {
			file_size += filelength(fd);
			((CMTT_INFO *)s)->f_have_tfm = TRUE;
		} else {
			((CMTT_INFO *)s)->f_have_tfm = FALSE;
		}
		tmp = CMTTF_ID;
	}
#endif
	if (tmp == JXL4_ID) {
		if (*name=='*')
			error(FILE_FAULT, "%s contains JXL4 file", name + 9);
		file_size = 64L;
		fp = fdopen(fd, "rb");
		fseek(fp, -64L, SEEK_END);
	}
	if (ask_pk_buf(file_size) == FAILURE)
		error(NO_MEMORY, "font file over: %s", to_font_name(name));
	if ((*((long far *)pk_buf_pointer->current) = tmp) == JXL4_ID) {
		jxl_info = (long far *)(pk_buf_pointer->current);
		for (i = 15; i > 0; i--)
			*(++jxl_info) = read_long(fp);
		if (read_long(fp) != 0x18B25L)
			error(FILE_FAULT, "JXL4 file %s", name);
		fclose(fp);
	}
#ifdef TTFONT
	else if (tmp == CMTTF_ID) {
		d = pk_buf_pointer->current + 4;
		for (i = sizeof(CMTT_INFO);  i > 0; i--) *d++ = *s++;
		if (fd > 0) {
			buf_load(fd, d, file_size - sizeof(CMTT_INFO) - 4);
			close(fd);
		}
	}
#endif
	else {
		buf_load(fd, pk_buf_pointer->current + 4, file_size - 4);
		close(fd);
	}
	pk = pk_buf_pointer->current;
	(HUGE_BUF *)pk_buf_pointer->current += file_size;
	RETURN(pk);
}

#if DEBUG & MEM_CHK
static void _pkbufchk(BUFFER *ptr)
	/*  pk-bufferaccess check
     */
{
	ENTER("_pkbufchk");

	if ((HUGE_BUF *)ptr < (HUGE_BUF *)pk_buf_pointer->start ||
		(HUGE_BUF *)ptr >= (HUGE_BUF *)pk_buf_pointer->end)
		error(MEMORY_FAULT, "pk-buffer");
	END();
}

#endif

#ifdef TTFONT
int load_tfm(char *name)
{
	int fd, len;
	char *tmp, *path, *dir, *ext = "tfm";

	/* tHg؂o */
	name += strlen(name);
	len = 0;
	while (--name >= tmp_buf) {
		if (*name == '\\' || *name == '/') break;
		len++;
		if (*name == '.') len = 0;
	}
	name++;
	name = strncpy(tmp_buf, name, len);
	path = tmp_buf + len;
	*path++ = '\0';

	dir = getenv("TEXFONTS");
	do {
		tmp = path;
		while (*dir != '\0' && *dir != ';')
			*tmp++ = *dir++;
		*tmp = '\0';
		if(!*path) continue;
		make_path(path, path, name, ext);
		if ((fd = openf(path, O_RDONLY | O_BINARY)) != FAILURE)
			return (fd);
	} while (*dir++);
	make_path(path, NULL, name, ext);
	error(WARNING, "Cannot find %s for TT Font", path);
	return (0);
}
#endif

BUFFER *
    load_pk(char *name)
	/*  namefontCꊇā@_pk_bufɓǂݍށD
     */
{
	ENTER("load_pK");

#if DEBUG & MEM_CHK
	_bufchk = _pkbufchk;
	/* buffer access check */
#endif

	RETURN(load_pk_data(name));
}

#if defined(TURBOC) || defined(MSC)
int file_read(int fd, BUFFER *ptr, int size)
	/* ZOgOɒڃ[h
     */
{
	union REGS inregs, outregs;
	struct SREGS segregs;

	ENTER("file_read");

	inregs.h.ah = 0x3f;
	segregs.ds = FP_SEG(ptr);
	inregs.x.dx = FP_OFF(ptr);
	inregs.x.cx = size;
	inregs.x.bx = fd;

	intdosx(&inregs, &outregs, &segregs);

	if (outregs.x.cflag)
		RETURN(FAILURE);
	else
		RETURN(outregs.x.ax);
}

#elif defined(x68k)
int file_read(int fd, BUFFER *buf, int size)
{
	return (read(fd, buf, size) == size ? 0 : FAILURE);
}

#endif

/* end of file : loadpk.c */
